const { EventEmitter } = require("events");
const { basename, dirname, join } = require("path");
const { EOL } = require("os");
const { isMaster } = require("cluster");

const isWindows = process.platform === "win32";

const maxAPI = require("./api");
const { CHILD_PROCESS_MESSAGES, ERROR_CODES, NAMED_PIPE_PREFIX, WAIT_FOR_CTRL_CONN_TIMEOUT } = require("./constants.js");
const { hrToMs, mysToMs, uniqID } = require("./utils");

// Add Node and NPM binaries to PATH when running on Windows
if (isWindows) {
	process.env.PATH = [
		dirname(process.env.N4M_NODE_PATH),
		dirname(process.env.N4M_NPM_PATH),
		process.env.PATH
	].join(";");
}

const exposedAPI = require("./exposed/max-api");

// we override process.exit in order to allow client scripts to smoothly exit by themselves if necessary
const exit = process.exit;

// Cluster Master Node Module Loader Monkey Patching for Watching feature
if (isMaster) {
	const Module = require("module");
	const _backupLoader = Module._load;

	const getRequiredModulePath = (modulePath, source) => {
		let fullPath;

		try {
			fullPath = require.resolve(modulePath);
		} catch (e) {
			fullPath = null;
			// to be expected when not a npm dependency
		}

		// System Modules, f.e. fs, path etc
		if (fullPath === modulePath) return null;

		if (fullPath) return fullPath;

		try {
			fullPath = join(dirname(source), modulePath);
			return Module._resolveFilename(fullPath);
		} catch (e) {
			return null;
		}
	};

	Module._load = (fr, context, isMain) => {
		const requiredPath = getRequiredModulePath(fr.slice(0), context.filename.slice(0));

		if (requiredPath) {
			process.send({
				type: CHILD_PROCESS_MESSAGES.WATCH_PATH,
				id: uniqID(),
				payload: {
					filePath: requiredPath
				}
			});
		}

		return _backupLoader.call(this, fr, context, isMain);
	};
}

// Indicator if we managed to load the user's script successfully
let loadedSuccessfully = false;

class NSProcessShell extends EventEmitter {

	constructor() {
		super();

		// this bound callbacks
		this.shutdown = this.shutdown.bind(this);
		this._handleUncaughtProcError = this._handleUncaughtProcError.bind(this);
		this._handleParentProcessMessage = this._handleParentProcessMessage.bind(this);

		// Overwrite process.exit with custom function call
		process.exit = this._callExit.bind(this);

		// Exception Handling
		process.on("uncaughtException", this._handleUncaughtProcError);
		process.on("unhandledRejection", this._handleUncaughtProcError);

		// listen for TERM signal .e.g. kill
		process.on("SIGTERM", this.shutdown);
		// listen for INT signal e.g. Ctrl-C
		process.on("SIGINT", this.shutdown);

		// Parent Process Message handling
		process.on("message", this._handleParentProcessMessage);

		this._controlSockPath = process.platform === "win32" ? `${NAMED_PIPE_PREFIX}${process.env.SCRIPT_ID}ctrl` : `/tmp/${process.env.SCRIPT_ID}_ctrl.sock`;
	}

	_handleControlSocketError(err) {
		process.send({
			type: CHILD_PROCESS_MESSAGES.UNCAUGHT_EXCEPTION,
			payload: {
				error: {
					message: err.message,
					details: err.details,
					code: err.code,
					stack: err.stack
				}
			}
		});
		this.shutdown(ERROR_CODES.UNCAUGHT_EXCEPTION);

	}

	_handleUncaughtProcError(err) {
		let type;
		let error;
		let code;

		if (loadedSuccessfully) {
			type = CHILD_PROCESS_MESSAGES.UNCAUGHT_EXCEPTION;
			code = ERROR_CODES.UNCAUGHT_EXCEPTION;
			error = {
				message: err.message,
				details: err.details,
				code: code,
				stack: err.stack
			};
		} else {
			type = CHILD_PROCESS_MESSAGES.SETUP_FAIL;
			code = ERROR_CODES.PROCESS_SETUP_FAIL;
			error = {
				message: "Unable to launch process",
				details: err.message,
				code: code,
				stack: err.stack
			};
		}
		process.send({
			type,
			payload: {
				error
			}
		});

		this.shutdown(code);
	}

	_callExit(code) {
		process.send({
			type: CHILD_PROCESS_MESSAGES.DISABLE_RESTART,
			id: uniqID()
		}, () => {
			this.shutdown(code);
		});
	}

	_getCPUUsage() {
		return new Promise((resolve, reject) => {

			const start = process.hrtime();
			const usage = process.cpuUsage();

			setTimeout(() => {
				const elapsedTime = hrToMs(process.hrtime(start));
				const elapsedUsage = process.cpuUsage(usage);

				const elapSys = mysToMs(elapsedUsage.system);
				const elapUser = mysToMs(elapsedUsage.user);

				return resolve({
					user: elapUser,
					system: elapSys,
					timeframe: elapsedTime,
					percentage: (elapUser + elapSys) / elapsedTime
				});
			}, 100);
		});
	}

	_handleParentProcessMessage(msg) {
		const { type, id } = msg;
		switch (type) {

			case CHILD_PROCESS_MESSAGES.PROCESS_STATUS: {
				this._getCPUUsage()
					.then((usage) => {
						process.send({
							type,
							id,
							payload: {
								cpuUsage: usage,
								cwd: process.cwd(),
								filename: basename(process.env.SCRIPT_PATH),
								memoryUsage: (() => {
									const { rss, heapUsed, heapTotal } = process.memoryUsage();
									return {
										rss,
										heapUsed,
										heapTotal
									};
								})(),
								nodePath: process.execPath,
								nodeVersion: process.version,
								uptime: Math.floor(process.uptime())
							}
						});
					});
				return;
			}
			case CHILD_PROCESS_MESSAGES.KILL: {
				process.send({
					type: type,
					id,
					payload: {}
				});
				this.shutdown();
				return;
			}
			default: {
				process.send({
					type: CHILD_PROCESS_MESSAGES.UNKNOWN_MESSAGE_WARNING,
					id: id,
					payload: {
						error: {
							message: `Unkown IPC Message ${type}`,
							code: ERROR_CODES.UNKNOWN_IPC_MESSAGE
						}
					}
				});
				return;
			}
		}
	}

	_onControlClientDisconnected() {
		if (this._shuttingDown) return; // expected

		const err = new Error("Remote control client disconnected");
		err.code = ERROR_CODES.CONTROL_SOCKET_DISCONNECTED;

		this._handleControlSocketError(err);
	}

	_onControlSocketClose() {
		if (this._shuttingDown) return; // expected

		const err = new Error("Control Server closed unexpectedly");
		err.code = ERROR_CODES.CONTROL_SOCKET_CLOSED;

		this._handleControlSocketError(err);
	}

	_onControlSocketError(sockErr) {
		if (this._shuttingDown) return; // expected

		const err = new Error("Control Server error encountered");
		err.code = ERROR_CODES.CONTROL_SOCKET_ERROR;
		err.details = sockErr.message;

		this._handleControlSocketError(err);
	}

	get shutdownFct() {
		return exposedAPI.__n4m_getRegisteredShutdownHook();
	}

	setup() {
		return new Promise(async (resolve, reject) => {
			let address;
			try {
				address = await maxAPI.listen();
			} catch (err) {
				return void reject(err);
			}

			maxAPI.once("error", this._onControlSocketError.bind(this));

			// We wait until the remote client has connected before we resolve to make
			// sure that EVERYTHING is in place/set up b4 the user defined script and other
			// tasks are executed. However, to make that happen we need so send the UP message b4
			let waitForConnectionTimeout = null;

			process.send({
				type: CHILD_PROCESS_MESSAGES.UP,
				payload: {
					controlSocket: address
				}
			});

			maxAPI.once("connected", () => {
				clearTimeout(waitForConnectionTimeout);

				// we consider any disconnection, closing etc an error!
				maxAPI.once("disconnected", this._onControlClientDisconnected.bind(this));
				maxAPI.once("close", this._onControlSocketClose.bind(this));

				return void resolve();
			});

			// Schedule Timeout to catch a never connecting Max end
			waitForConnectionTimeout = setTimeout(() => {
				maxAPI.removeAllListeners();

				const err = new Error("Control client connection timed out.");
				err.details = `The Max end never connected within ${WAIT_FOR_CTRL_CONN_TIMEOUT}ms`;
				err.code = ERROR_CODES.CONTROL_SOCKET_TIMEOUT;

				return void reject(err);

			}, WAIT_FOR_CTRL_CONN_TIMEOUT);
		});
	}

	async shutdown(code = 0) {
		this._shuttingDown = true;

		// Close any potential lingering debugging connection
		if (process.debugPort) require("inspector").close();

		let returnVal;
		if (this._shutdownFct) returnVal = this._shutdownFct();

		if (!returnVal || !(returnVal instanceof Promise)) returnVal = Promise.resolve();

		try {
			await returnVal;
			await maxAPI.destroy();
		} catch (err) {
			// ignore err here as we are dismantling the process anyway
		} finally {
			exit(code);
		}
	}
}

const main = async () => {
	if (isMaster) {
		// Cluster Master with full MaxAPI and Process Handling support

		// The Shell to Handle the Process
		let runnerShell = null;
		try {
			runnerShell = new NSProcessShell();
			await runnerShell.setup();

			// write initial newline char
			process.stdout.write(EOL);
			process.stderr.write(EOL);

			// Kick off things and load user script
			require(process.env.SCRIPT_PATH);
			loadedSuccessfully = true;
		} catch (err) {
			process.send({
				type: CHILD_PROCESS_MESSAGES.SETUP_FAIL,
				payload: {
					error: {
						message: "Unable to launch process",
						details: err.message,
						code: ERROR_CODES.PROCESS_SETUP_FAIL,
						stack: err.stack
					}
				}
			});
			runnerShell.shutdown(ERROR_CODES.PROCESS_SETUP_FAIL);
		}
	} else {
		// Cluster Worker
		// No special setup of MaxAPI for now.

		// Kick off things and load user script
		require(process.env.SCRIPT_PATH);
		loadedSuccessfully = true;
	}
};

main();
